@@ -31,7 +31,20 @@ print.htmlwidget <- function(x, ..., view = interactive()) {
 
 #' @export
 print.suppress_viewer <- function(x, ..., view = interactive()) {
-  html_print(htmltools::as.tags(x, standalone=TRUE), viewer = if (view) browseURL)
+
+  viewer <- if (view) {
+    if (isTRUE(x$sizingPolicy$browser$external)) {
+      browseURL
+    } else if (!is.null(getOption("page_viewer"))) {
+      function(url) getOption("page_viewer")(url)
+    } else {
+      browseURL
+    }
+  } else {
+    NULL
+  }
+
+  html_print(htmltools::as.tags(x, standalone=TRUE), viewer = viewer)
   invisible(x)
 }
 
@@ -76,15 +89,23 @@ appendContent <- function(x, ...) {
 #'
 #' @param x An HTML Widget object
 #' @param jsCode Character vector containing JavaScript code (see Details)
+#' @param data An additional argument to pass to the \code{jsCode} function.
+#'   This can be any R object that can be serialized to JSON. If you have
+#'   multiple objects to pass to the function, use a named list.
 #' @return The modified widget object
 #'
 #' @details The \code{jsCode} parameter must be a valid JavaScript expression
 #'   that returns a function.
 #'
-#'   The function will be invoked with two arguments: the first is the widget's
+#'   The function will be invoked with three arguments: the first is the widget's
 #'   main HTML element, and the second is the data to be rendered (the \code{x}
-#'   parameter in \code{createWidget}). When the function is invoked, the
-#'   \code{this} will be the widget instance object.
+#'   parameter in \code{createWidget}). The third argument is the JavaScript
+#'   equivalent of the R object passed into \code{onRender} as the \code{data}
+#'   argument; this is an easy way to transfer e.g. data frames without having
+#'   to manually do the JSON encoding.
+#'
+#'   When the function is invoked, the \code{this} keyword will refer to the
+#'   widget instance object.
 #'
 #' @seealso \code{\link{onStaticRenderComplete}}, for writing custom JavaScript
 #'   that involves multiple widgets.
@@ -93,6 +114,9 @@ appendContent <- function(x, ...) {
 #' \dontrun{
 #' library(leaflet)
 #'
+#' # This example uses browser geolocation. RStudio users:
+#' # this won't work in the Viewer pane; try popping it
+#' # out into your system web browser.
 #' leaflet() %>% addTiles() %>%
 #'   onRender("
 #'     function(el, x) {
@@ -100,21 +124,46 @@ appendContent <- function(x, ...) {
 #'       this.locate({setView: true});
 #'     }
 #'   ")
+#'
+#'
+#' # This example shows how you can make an R data frame available
+#' # to your JavaScript code.
+#'
+#' meh <- "&#x1F610;";
+#' yikes <- "&#x1F628;";
+#'
+#' df <- data.frame(
+#'   lng = quakes$long,
+#'   lat = quakes$lat,
+#'   html = ifelse(quakes$mag < 5.5, meh, yikes),
+#'   stringsAsFactors = FALSE
+#' )
+#'
+#' leaflet() %>% addTiles() %>%
+#'   fitBounds(min(df$lng), min(df$lat), max(df$lng), max(df$lat)) %>%
+#'   onRender("
+#'     function(el, x, data) {
+#'       for (var i = 0; i < data.lng.length; i++) {
+#'         var icon = L.divIcon({className: '', html: data.html[i]});
+#'         L.marker([data.lat[i], data.lng[i]], {icon: icon}).addTo(this);
+#'       }
+#'     }
+#'   ", data = df)
 #' }
 #'
 #' @export
-onRender <- function(x, jsCode) {
-  addHook(x, "render", jsCode)
+onRender <- function(x, jsCode, data = NULL) {
+  addHook(x, "render", jsCode, data)
 }
 
-addHook <- function(x, hookName, jsCode) {
+addHook <- function(x, hookName, jsCode, data = NULL) {
   if (length(jsCode) == 0)
     return(x)
 
   if (length(jsCode) > 1)
     jsCode <- paste(jsCode, collapse = "\n")
 
-  x$jsHooks[[hookName]] <- c(x$jsHooks[[hookName]], list(jsCode))
+  x$jsHooks[[hookName]] <- c(x$jsHooks[[hookName]], list(list(code = jsCode, data = data)))
   x
 }
 
@@ -126,7 +175,7 @@ toHTML <- function(x, standalone = FALSE, knitrOptions = NULL) {
   if (!is.null(x$elementId))
     id <- x$elementId
   else
-    id <- paste("htmlwidget", as.integer(stats::runif(1, 1, 10000)), sep="-")
+    id <- paste("htmlwidget", createWidgetId(), sep="-")
 
   w <- validateCssUnit(sizeInfo$width)
   h <- validateCssUnit(sizeInfo$height)
@@ -228,7 +277,7 @@ widget_data <- function(x, id, ...){
 #' the specified package.
 #'
 #' For additional details on developing widgets, see package vignettes:
-#' \code{vignette("htmlwidgets-intro", package = "htmlwidgets")}.
+#' \code{vignette("develop_intro", package = "htmlwidgets")}.
 #'
 #' @param name Widget name (should match the base name of the YAML and
 #'   JavaScript files used to implement the widget)
@@ -307,6 +356,8 @@ createWidget <- function(name,
 #' for the output
 #' @param outputFunction Shiny output function corresponding to this render
 #'   function.
+#' @param reportSize Should the widget's container size be reported in the
+#'   shiny session's client data?
 #' @param expr An expression that generates an HTML widget
 #' @param env The environment in which to evaluate \code{expr}.
 #' @param quoted Is \code{expr} a quoted expression (with \code{quote()})? This
@@ -334,13 +385,13 @@ createWidget <- function(name,
 #'
 #' @export
 shinyWidgetOutput <- function(outputId, name, width, height, package = name,
-                              inline = FALSE) {
+                              inline = FALSE, reportSize = FALSE) {
 
   checkShinyVersion()
   # generate html
   html <- htmltools::tagList(
     widget_html(name, package, id = outputId,
-      class = paste(name, "html-widget html-widget-output"),
+      class = paste0(name, " html-widget html-widget-output", if (reportSize) " shiny-report-size"),
       style = sprintf("width:%s; height:%s; %s",
         htmltools::validateCssUnit(width),
         htmltools::validateCssUnit(height),
@@ -422,3 +473,6 @@ createPayload <- function(instance){
   list(x = x, evals = JSEvals(x), jsHooks = instance$jsHooks)
 }
 
+# package globals
+.globals <- new.env(parent = emptyenv())
+
